class DefaultHandler:

    @staticmethod
    def origin2flow(request_data):
        return request_data

    @staticmethod
    def flow2origin(flow_data):
        return flow_data

    @staticmethod
    def origin2string(request_data):
        return request_data



from .plugin_loader import manifest, get_plugin_storage
from .plugin_manager import PluginManager
from .status_bar import ClickableStatusText, ImageMenuItem, TextMenuItem, LinkMenuItem, Selector, PlainText



import binascii


class DefaultHandler:

    @staticmethod
    def origin2flow(request_data):
        _data = binascii.b2a_base64(request_data).decode('utf-8')
        return _data

    @staticmethod
    def flow2origin(flow_data):
        _data = binascii.a2b_base64(flow_data)
        return _data

    @staticmethod
    def origin2string(request_data):
        _data = binascii.b2a_base64(request_data).decode('utf-8')
        return _data



from urllib.parse import urlencode, parse_qs


class FormHandler:

    @staticmethod
    def origin2flow(request_data):
        params = parse_qs(request_data.decode('utf-8'), keep_blank_values=True)
        _data = {k:v[0] for k,v in params.items()}
        return _data

    @staticmethod
    def flow2origin(flow_data):
        _data = urlencode(flow_data).encode()
        return _data

    @staticmethod
    def origin2string(request_data):
        return request_data.decode('utf-8')



from lyrebird.base_server import StaticServer
from . import plugin_loader
from lyrebird import application
from lyrebird.checker.on_request import OnRequestHandler
from lyrebird.checker.on_response import OnResponseHandler
from lyrebird.checker.on_request_upstream import OnRequestUpstreamHandler
from lyrebird.checker.on_response_upstream import OnResponseUpstreamHandler
from flask import Blueprint, send_file, request
from lyrebird.log import get_logger
from types import FunctionType
import traceback

logger = get_logger()


class PluginManager(StaticServer):

    def __init__(self):
        self.plugin_path_list = []
        self.plugins = {}

    def reload(self):
        self.plugins = {}
        self.plugins.update(plugin_loader.load_all_from_ep())
        for plugin_path in self.plugin_path_list:
            try:
                plugin = plugin_loader.load_from_path(plugin_path)
                logger.info(f'Load plugin from path success: {plugin.project_name}')
            except:
                logger.error(f'Load plugin failed. Skip plugin : {plugin_path}\n{traceback.format_exc()}')
                continue
            self.plugins[plugin.project_name] = plugin
        self.setup_plugins()

    def setup_plugins(self):
        application.config['extension.plugin.enable'] = []
        # TODO register plugins pb
        for p_name, plugin in self.plugins.items():
            application.config['extension.plugin.enable'].append(p_name)
            view_static_folder_name = plugin.manifest.view[0]
            view_entry_file = plugin.manifest.view[1]
            plugin_static_folder = f"{plugin.location}/{view_static_folder_name}"

            # Create new blueprint for each plugin
            _bp = Blueprint(
                f'plugins_{p_name}',
                f'plugins_{p_name}',
                url_prefix=f'/plugins/{p_name}',
                static_folder=plugin_static_folder)

            # Add plugin main view to flask blueprint
            _view = plugin.manifest.view
            _index_file_path = f"{plugin.location}/{view_static_folder_name}/{view_entry_file}"
            _bp.add_url_rule('/', view_func=IndexPageViewFunc(p_name, _index_file_path))

            @_bp.after_request
            def print_plugin_api(response):
                lyrebird_info = response.headers.get('lyrebird', default='')
                logger.info(
                    f'[On plugin API]{response.status_code} {lyrebird_info} {request.method} {request.url[:100]}')
                return response

            # Add API to blureprint
            for api in plugin.manifest.api:
                rule = api[0]
                view_func = api[1]
                if not isinstance(view_func, FunctionType):
                    logger.error(f'View func not callable. Skip API: {api}. Plugin: {p_name}')
                    continue
                if len(api) > 2:
                    methods = api[2]
                else:
                    methods = ['GET']
                _bp.add_url_rule(rule, view_func=view_func, methods=methods)

            # Register blueprint
            mock_service = application.server['mock']
            mock_service.app.register_blueprint(_bp)

            # Load background task
            for task in plugin.manifest.background:
                application.server['task'].add_task(task[0], task[1])

            # Subscribe event linstener
            event_service = application.server['event']
            for event_option in plugin.manifest.event:
                channel = event_option[0]
                callback_func = event_option[1]
                event_service.subscribe(channel, callback_func)

            # Subscribe handler on request
            for handler in plugin.manifest.on_request:
                OnRequestHandler.register({
                    'name': handler[0],
                    'func': handler[1],
                    'rules': handler[2] if len(handler) > 2 else None,
                    'rank': handler[3] if len(handler) > 3 and isinstance(handler[3], (int, float)) else 0,
                    'modify_request_body': handler[4] if len(handler) > 4 else True
                })

            # Subscribe handler on response
            for handler in plugin.manifest.on_response:
                OnResponseHandler.register({
                    'name': handler[0],
                    'func': handler[1],
                    'rules': handler[2] if len(handler) > 2 else None,
                    'rank': handler[3] if len(handler) > 3 and isinstance(handler[3], (int, float)) else 0
                })

            # Subscribe handler on proxy request
            for handler in plugin.manifest.on_request_upstream:
                OnRequestUpstreamHandler.register({
                    'name': handler[0],
                    'func': handler[1],
                    'rules': handler[2] if len(handler) > 2 else None,
                    'rank': handler[3] if len(handler) > 3 and isinstance(handler[3], (int, float)) else 0,
                    'modify_request_body': handler[4] if len(handler) > 4 else True
                })

            # Subscribe handler on proxy response
            for handler in plugin.manifest.on_response_upstream:
                OnResponseUpstreamHandler.register({
                    'name': handler[0],
                    'func': handler[1],
                    'rules': handler[2] if len(handler) > 2 else None,
                    'rank': handler[3] if len(handler) > 3 and isinstance(handler[3], (int, float)) else 0
                })

            for status_item in plugin.manifest.status:
                plugin.status.append(status_item())


class IndexPageViewFunc:

    def __init__(self, name, index_file_path):
        self.__name__ = name
        self.index_file_path = index_file_path

    def __call__(self):
        return send_file(self.index_file_path)



from lyrebird.log import get_logger
from .server import serve, publish_init_status
from lyrebird.base_server import ProcessServer

logger = get_logger()


class ExtraMockServer(ProcessServer):

    def run(self, msg_queue, config, log_queue, *args, **kwargs):
        publish_init_status(msg_queue, 'READY')
        serve(msg_queue, config, log_queue, *args, **kwargs)



from flask import Blueprint, request
from flask_restful import Api
from .common import Status, WorkMode, Manifest, DiffMode, Render
from .flow import Flow, FlowList
from .mock import MockData, MockGroup, ActivatedMockGroup, MockGroupByName, MockDataLabel, TreeView, OpenNodes
from .config import Conf
from .menu import Menu
from .notice import Notice
from .checker import Checker
from .event import Event, EventExport, Channel, EventFileInfo
from .conflict_check import ConflictCheck, ActivatedDataConflictCheck
from .mock_editor import Cut, Copy, Paste, Duplicate
from .qrcode import Qrcode
from .search import SearchMockDataByName
from .bandwidth import Bandwidth, BandwidthTemplates
from .status_bar import StatusBar
from .snapshot import SnapshotImport, SnapshotExport, Snapshot
from lyrebird.log import get_logger
from lyrebird import application


logger = get_logger()


api = Blueprint('api', __name__, url_prefix='/api')
api_source = Api(api, errors=Exception)


@api.after_request
def after_request(response):
    """
    输出每条请求概要信息
    """
    lyrebird_info = response.headers.get('lyrebird', default='')
    logger.info(f'[On API]{response.status_code} {lyrebird_info} {request.method} {request.url[:100]}')
    return response


api_source.add_resource(Status, '/status')
api_source.add_resource(Manifest, '/manifest')
api_source.add_resource(Flow, '/flow/<string:id>')
api_source.add_resource(FlowList, '/flow', '/flow/<string:action>')
api_source.add_resource(Cut, '/cut/<string:_id>')
api_source.add_resource(Copy, '/copy/<string:_id>')
api_source.add_resource(Paste, '/paste/<string:_id>')
api_source.add_resource(Duplicate, '/duplicate/<string:_id>')
api_source.add_resource(TreeView, '/tree')
api_source.add_resource(OpenNodes, '/tree/open-nodes')
api_source.add_resource(MockGroup, '/group', '/group/<string:group_id>', '/group/label/<string:label>')
api_source.add_resource(MockData, '/data', '/data/<string:_id>')
api_source.add_resource(SearchMockDataByName, '/search/group/name/', '/search/group/name/<string:search_str>')
api_source.add_resource(ConflictCheck, '/conflict/id/<string:group_id>')
api_source.add_resource(ActivatedDataConflictCheck, '/conflict/activated')
api_source.add_resource(ActivatedMockGroup, '/mock/activated', '/mock/<string:group_id>/<string:action>')
api_source.add_resource(MockGroupByName, '/mock_by_name')
api_source.add_resource(MockDataLabel, '/label')
api_source.add_resource(Qrcode, '/qrcode')
api_source.add_resource(Conf, '/conf')
api_source.add_resource(WorkMode, '/mode', '/mode/<string:mode>')
api_source.add_resource(DiffMode, '/diffmode')
api_source.add_resource(Render, '/render')
api_source.add_resource(Menu, '/menu')
api_source.add_resource(Notice, '/notice')
api_source.add_resource(Checker, '/checker', '/checker/<string:checker_id>', '/checker/search')
api_source.add_resource(Bandwidth, '/bandwidth')
api_source.add_resource(BandwidthTemplates, '/bandwidth_templates')
api_source.add_resource(SnapshotImport, '/snapshot/import')
api_source.add_resource(SnapshotExport, '/snapshot/export/event', '/snapshot/export/<string:group_id>')
api_source.add_resource(Snapshot, '/snapshot/<string:snapshot_id>')
api_source.add_resource(
    Event,
    '/event',
    '/event/page/<int:page>',
    '/event/page/<int:page>/search',
    '/event/page/<int:page>/search/<string:search_str>',
    '/event/id/<string:event_id>',
    '/event/id/<string:event_id>/search',
    '/event/id/<string:event_id>/search/<string:search_str>',
    '/event/search',
    '/event/search/<string:search_str>',
    '/event/<string:channel>',
    '/event/<string:channel>/search',
    '/event/<string:channel>/search/<string:search_str>',
    '/event/<string:channel>/page/<int:page>',
    '/event/<string:channel>/page/<int:page>/search',
    '/event/<string:channel>/page/<int:page>/search/<string:search_str>',
    '/event/<string:channel>/id/<string:event_id>',
    '/event/<string:channel>/id/<string:event_id>/search',
    '/event/<string:channel>/id/<string:event_id>/search/<string:search_str>'
)
api_source.add_resource(EventExport, '/event/export', '/event/export/<string:event_id>')
api_source.add_resource(Channel, '/channel', '/channel/<string:mode>')
api_source.add_resource(StatusBar, '/statusbar', '/statusbar/<string:item_id>')
api_source.add_resource(EventFileInfo, '/event/fileinfo')



import json
from lyrebird.log import get_logger


logger = get_logger()
class DuplicateHeaderKeyHandler:

    @staticmethod
    def set_origin_header(origin_header, flow_header):
        if not flow_header:
            return

        raw_header = DuplicateHeaderKeyHandler.flow2origin(flow_header)
        header_dict = dict()
        for key, value in raw_header:
            if key.lower() != 'set-cookie':
                continue
            if key not in header_dict:
                header_dict[key] = []
            header_dict[key].append(value)
        
        for key, value_list in header_dict.items():
            origin_header.remove(key)
            for value in value_list:
                origin_header.add_header(key, value)

    @staticmethod
    def origin2flow(raw_header):
        if not raw_header:
            return
        temp_header_dict = dict()
        for key, value in raw_header.items():
            if key not in temp_header_dict:
                temp_header_dict[key] = []
            temp_header_dict[key].append(value)
        
        header_dict = dict()
        for key, value_list in temp_header_dict.items():
            if key.lower() in trans_key_handlers:
                encode_func = trans_key_handlers.get(key.lower()).get('encode')
                encode_func(header_dict, key, value_list)
            else:
                header_dict[key] = value_list[0]
        
        return header_dict

    @staticmethod
    def flow2origin(header_dict):
        if not header_dict:
            return
        header_list = []
        for key, value in header_dict.items():
            if key.lower() in trans_key_handlers:
                decode_func = trans_key_handlers.get(key.lower()).get('decode')
                decode_func(header_list, key, value)
            else:
                header_list.append((key, value))
        return header_list
    
    @staticmethod
    def _to_json_string(header_dict, key, value_list):
        if len(value_list) == 1:
            header_dict[key] = value_list[0]
        else:
            header_dict[key] = json.dumps(value_list)
    
    @staticmethod
    def _to_json_object(header_list, key, value):
        try:
            raw_value = json.loads(value)
        except:
            # as a string
            raw_value = value
        if type(raw_value) == list:
            for value in raw_value:
                header_list.append((key, value))
        else:
            header_list.append((key, raw_value))


trans_key_handlers = {
    'set-cookie': {
        'encode': DuplicateHeaderKeyHandler._to_json_string,
        'decode': DuplicateHeaderKeyHandler._to_json_object
    }
}



import webbrowser

from flask import jsonify
from functools import reduce

"""
Lyrebird context

"""


def make_ok_response(**kwargs):
    ok_resp = {
        "code": 1000,
        "message": "success"
    }
    ok_resp.update(kwargs)
    return jsonify(ok_resp)


def make_fail_response(msg, **kwargs):
    fail_resp = {
        "code": 3000,
        "message": msg
    }
    fail_resp.update(kwargs)
    return jsonify(fail_resp)


_cm = None
_src = None


server = {}
plugins = {}
active_menu = {}


notice = None
checkers = {}

on_request = []
on_response = []
on_request_upstream = []
on_response_upstream = []

encoder = []
decoder = []

labels = None

encoders_decoders = None


def start_server_without_mock_and_log():
    for name in server:
        if name == 'mock' or name == 'log':
            continue
        server[name].start()


def start_mock_server():
    server['mock'].start()


def start_log_server():
    server['log'].start()


def start_server():
    for name in server:
        server[name].start()


def stop_server():
    for name in server:
        server[name].stop()


class ConfigProxy:

    def get(self, k, default=None):
        return _cm.config.get(k, default)

    def __setitem__(self, k, v):
        _cm.config[k] = v

    def __getitem__(self, k):
        return _cm.config[k]

    def raw(self):
        return _cm.config


config = ConfigProxy()

# statistics reporter
reporter = None


def root_dir():
    if _cm:
        return _cm.ROOT


NO_BROSWER = False

# --------- Lyrebird status ---------
'''
[INITING] --> /run lyrebird main method/  --> [READY]
              /start services          / 
              /extra mock server ready /
              /mitm proxy server ready /
'''

# Lyrebird status contains: 'READY' and 'INITING'
status = 'INITING'

# mitm proxy check point will be added if args set useing mitm server (--no-mitm False)
status_checkpoints = {
    'main': False,
    'extra_mock': False
}


def status_listener(event):
    '''
    event example
    event = {
        'channel': 'system',
        'system': {
            'action': 'init_module',
            'status': 'READY',
            'module': 'main'
        }
    }
    '''

    system = event.get('system')
    if not system:
        return

    action = system.get('action')
    if action != 'init_module':
        return

    module = system.get('module')
    if module not in status_checkpoints:
        return

    module_status = system.get('status')
    if module_status == 'READY':
        status_checkpoints[module] = True

    is_all_status_checkpoints_ok = reduce(lambda x, y: x and y, status_checkpoints.values())
    if is_all_status_checkpoints_ok:
        # Set global status
        global status
        status = 'READY'

        server['event'].publish('system', {'action': 'init_module', 'status': 'READY', 'module': 'all'})

        # auto open web browser
        if not NO_BROSWER:
            webbrowser.open(f'http://localhost:{config["mock.port"]}')

def process_status_listener():
    server['event'].subscribe('system', status_listener)


def status_ready():
    server['event'].publish('system', {
        'system': {'action': 'init_module', 'status': 'READY', 'module': 'main'}
    })



STATUS_INFINITE_PROXY_LOOP = '901 INFINITE PROXY LOOP'
STATUS_CODE_INFINITE_PROXY_LOOP = 901

STATUS_CAN_NOT_HANDLE_REQUEST = '904 CAN NOT HANDLE REQUEST'
STATUS_CODE_CAN_NOT_HANDLE_REQUEST = 904

STATUS_PROXY_ERROR = '905 PROXY ERROR'
STATUS_CODE_PROXY_ERROR = 905



from flask import jsonify
from . import cache
from .dm import DataManager
from .dm.file_data_adapter import data_adapter
from flask_socketio import SocketIO
from importlib import machinery
from pathlib import Path
import codecs
import json
import os
from lyrebird import application as app
from lyrebird.log import get_logger
import traceback
import time


"""
Mock server context

"""

logger = get_logger()


class DataManagerInitException(Exception):
    pass

class Mode:
    NORMAL = 'normal'
    RECORD = 'record'

    @staticmethod
    def contains(val):
        if val == Mode.NORMAL or val == Mode.RECORD:
            return True
        else:
            return False


class MockMode:
    NORMAL = 'normal'
    MULTIPLE = 'multiple'

    @staticmethod
    def contains(val):
        return val == MockMode.NORMAL or val == MockMode.MULTIPLE


class Application:

    def __init__(self):
        self._conf = None
        # todo 使用内存中的List存储请求，应可支持切换redis
        self.cache = cache.get_cache()
        self.work_mode = Mode.NORMAL
        self.is_diff_mode = MockMode.NORMAL
        self.filters = []
        self.selected_filter = {}
        self.data_manager = DataManager()
        # SocketIO
        self.socket_io: SocketIO = None
        self.conf_manager = None
        self.current_conf_name = None

    @property
    def event_bus(self):
        return app.server.get('event')

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self, _conf):
        self._conf = _conf

        if _conf.get('mock.data'):
            self.init_datamanager(_conf.get('mock.data'))

        if _conf.get('mock.mode') == MockMode.MULTIPLE:
            self.is_diff_mode = MockMode.MULTIPLE

        if _conf.get('inspector.filters'):
            self.filters.extend(_conf.get('inspector.filters'))

        default_filter = _conf.get('inspector.default_filter')
        for f in self.filters:
            if f['name'] == default_filter:
                self.selected_filter = f
                break

    def init_datamanager(self, uri):
        adapter_cls = data_adapter

        path = Path(uri).expanduser().absolute()

        if path.exists() and path.is_file():
            logger.warning(f'Loading custom data-adapter from {str(path)} .')
            try:
                loader = machinery.SourceFileLoader(path.stem, str(path))
                adapter_module = loader.load_module()
                adapter_cls = adapter_module.data_adapter
            except Exception:
                logger.error(f'Loading custom data-adapter from {str(path)} failed!\n{traceback.format_exc()}')
                raise DataManagerInitException
        else:
            logger.warning(f'Loading file data-adapter from {str(path)} .')
        
        self.data_manager.set_adapter(adapter_cls)
        self.data_manager.set_root(uri)

    def save(self):
        DEFAULT_CONF = os.path.join(
            os.path.join(os.path.expanduser('~'), '.lyrebird'), 'conf.json')
        # TODO Lyrebird 的 conf.json 应该有恢复机制或补救措施。
        with codecs.open(DEFAULT_CONF, 'w', 'utf-8') as f:
            f.write(json.dumps(self._conf, ensure_ascii=False, indent=4))

    def get_and_update_selected_filter_by_name(self, filter_name):
        old_filter = self.selected_filter
        if not filter_name:
            self.selected_filter = {}
        elif not old_filter or filter_name != old_filter['name']:
            for item in application.filters:
                if item['name'] == filter_name:
                    self.selected_filter = item
                    break
        app.config['inspector.default_filter'] = filter_name
        return self.selected_filter


"""
SocketIO emit interval
Because of iview table has render preformance problem
We need to limit render time
"""
EMIT_INTERVAL = 1
last_emit_time = {}

application = Application()

def make_ok_response(**kwargs):
    ok_resp = {
        "code": 1000,
        "message": "success"
    }
    ok_resp.update(kwargs)
    return jsonify(ok_resp)


def make_fail_response(msg, code=3000, **kwargs):
    fail_resp = {
        "code": code,
        "message": msg
    }
    fail_resp.update(kwargs)
    return jsonify(fail_resp)


def emit(event, *args, **kwargs):
    now = time.time()
    last_push_time = last_emit_time.get(event, 0)
    if (now - last_push_time) > EMIT_INTERVAL:
        application.socket_io.emit(event, *args, **kwargs)
        last_emit_time[event] = now



"""
Base threading server class
"""

from threading import Thread
from multiprocessing import Process, Queue
from lyrebird import application


service_msg_queue = Queue()


class ProcessServer:
    def __init__(self):
        self.server_process = None
        self.running = False
        self.name = None
        self.event_thread = None
        self.args = []
        self.kwargs = {}

    def run(self, msg_queue, config, log_queue, *args, **kwargs):
        '''
        msg_queue
        message queue for process server and main process

        #1. Send event to main process,
        {
            "type": "event",
            "channel": "",
            "content": {}
        }

        #2. Send message to frontend
        support channel: msgSuccess msgInfo msgError
        {
            "type": "ws",
            "channel": "",
            "content": ""
        }

        config
        lyrebird config dict

        log_queue
        send log msg to logger process
        '''
        pass

    def start(self):
        if self.running:
            return

        global service_msg_queue
        config = application.config.raw()
        logger_queue = application.server['log'].queue
        self.server_process = Process(group=None, target=self.run,
                                      args=[service_msg_queue, config, logger_queue, self.args],
                                      kwargs=self.kwargs,
                                      daemon=True)
        self.server_process.start()
        self.running = True

    def stop(self):
        if self.server_process:
            self.server_process.terminate()
            self.server_process = None


class ThreadServer:

    def __init__(self):
        self.server_thread = None
        self.running = False
        self.name = None

    def start(self, *args, **kwargs):
        if self.running:
            return
        self.running = True
        self.server_thread = Thread(target=self.run, name=self.name, args=args, kwargs=kwargs)
        self.server_thread.start()

    def stop(self):
        self.running = False
        # TODO terminate self.server_thread

    def run(self):
        """
        Server main function
        """
        pass


class StaticServer:

    def start(self, *args, **kwargs):
        pass

    def stop(self):
        pass


class MultiProcessServerMessageDispatcher(ThreadServer):

    def run(self):
        global service_msg_queue
        emit = application.server['mock'].socket_io.emit
        publish = application.server['event'].publish

        while True:
            msg = service_msg_queue.get()
            type = msg.get('type')
            if type == 'event':
                channel = msg.get('channel')
                event = msg.get('content')
                if channel and event:
                    publish(channel, event)
            elif type == 'ws':
                ws_channel = msg.get('channel')
                ws_msg = msg.get('content', '')
                if ws_channel:
                    emit(ws_channel, ws_msg)
            else:
                pass



import re
import os
import json
import math
import time
import socket
import tarfile
import requests
import datetime
import netifaces
import traceback
from pathlib import Path
from jinja2 import Template, StrictUndefined
from jinja2.exceptions import UndefinedError, TemplateSyntaxError
from contextlib import closing
from urllib.parse import unquote
from lyrebird.log import get_logger
from lyrebird.application import config

logger = get_logger()


def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


def convert_size_to_byte(size_str: str):
    size_str = size_str.strip().upper()
    match = re.match(r'^(\d+\.?\d*)\s*([KMGTPEZY]?[B])$', size_str)
    if not match:
        logger.warning(f'Invalid size string: {size_str}')
        return
    size = float(match.group(1))
    unit = match.group(2)
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = size_name.index(unit)
    size_bytes = int(size * (1024 ** i))
    return size_bytes


def convert_time(duration):
    if duration < 1:
        return str(round(duration * 1000)) + 'ms'
    else:
        return str(round(duration, 2)) + 's'


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        print(f'{method.__name__} execution time {(te-ts)*1000}')
        return result
    return timed


def is_port_in_use(port, host='127.0.0.1'):
    sock = None
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(1)
        sock.connect((host, int(port)))
        return True
    except socket.error:
        return False
    finally:
        if sock:
            sock.close()


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def get_ip():
    """
    Get local ip from socket connection

    :return: IP Addr string
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 80))
    return s.getsockname()[0]


def get_interface_ipv4():
    ipv4_list = []
    interface_list = netifaces.interfaces()
    for interface_name in interface_list:
        if interface_name == 'lo0':
            continue
        interface_dict = netifaces.ifaddresses(interface_name)
        interface = interface_dict.get(netifaces.AF_INET)
        if not interface:
            continue
        for alias in interface:
            ipv4_list.append({
                'address': alias['addr'],
                # windows keyerror 'netmask'
                # https://github.com/Meituan-Dianping/lyrebird/issues/665
                'netmask': alias.get('netmask', ''),
                'interface': interface_name
            })
    return ipv4_list


def compress_tar(input_path, output_path, suffix=None):
    current_path = Path.cwd()
    input_path = Path(input_path).expanduser().absolute().resolve()
    output_path = Path(output_path).expanduser().absolute().resolve()
    output_file = f'{output_path}{suffix}' if suffix else output_path

    os.chdir(input_path)
    # If not chdir, the directory in the compressed file will start from the root directory
    tar = tarfile.open(output_file, 'w:gz')
    for root, dirs, files in os.walk(input_path):
        for f in files:
            tar.add(f, recursive=False)
    tar.close()
    os.chdir(current_path)
    return output_file


def decompress_tar(input_path, output_path=None):
    input_path = Path(input_path).expanduser().absolute().resolve()
    if not output_path:
        filename = input_path.stem if input_path.suffix else f'{input_path.name}-decompress'
        output_path = input_path.parent / filename

    tf = tarfile.open(str(input_path))
    tf.extractall(str(output_path))
    tf.close()
    return output_path


def download(link, input_path):
    resp = requests.get(link, stream=True)
    with open(input_path, 'wb') as f:
        for chunck in resp.iter_content():
            f.write(chunck)


def handle_jinja2_tojson_by_config(data):
    config_value_tojson_key = config.get('config.value.tojsonKey')
    data_with_tojson = data
    for tojson_key in config_value_tojson_key:

        # EXAMPLE
        # response_data = `"key1":"value1","key2":"{{config.get('model.id')}}","key3":"value3"`
        # target_match_data = `"{{config.get('model.id')}}"`
        # Divide target_match_data into three parts `"{{` and `config.get('model.id')` and `}}"`
        # In the second part, `model.id` is a matching rule from Lyrebird configuration
        # The final return response_data is `"key1":"value1","key2":{{config.get('model.id') | tojson}},"key3":"value3"`

        pattern = '[^:]*' + tojson_key + '[^,]*'
        # The format of the group is required
        pattern_group = '(' + pattern + ')'
        data_with_tojson = re.sub('("{{)'+pattern_group+'(}}")', r'{{\2 | tojson}}', data_with_tojson)
    return data_with_tojson


def handle_jinja2_keywords(data, params=None):
    '''
    Jinja2 will throw an exception when dealing with unexpected left brackets, but right brackets will not
    So only left brackets need to be handled

    Handle 3 kinds of unexpected left brackets:
    1. More than 2 big brackets              `{{{ }}}` `{{# #}}` `{{% %}}`
    2. Mismatched keyword                    `{{{` `{{#` `{{%`
    3. Unknown arguments between {{ and }}   `{{unknown}}`

    Convert unexpected brackets into presentable string in Jinja2, such as `var` -> `{{ 'var' }}`
    The unexpected left brackets above will be convert into:
    `{{#`          ->  `{{ '{{#' }}`
    `{{unknown}}`  ->  `{{ '{{' }}unknown{{ '}}' }}`
    '''

    keywords_pair = {
        '{{': '}}',
        '{%': '%}',
        '{#': '#}'
    }

    # split by multiple `{` followed by `{` or `#`` or `%`, such as `{{`, `{{{`, `{{{{`, `{#`, `{{#`
    # EXAMPLE
    # data = '{{ip}} {{ip {{{ip'
    # item_list = ['', '{{', 'ip}} ', '{{', 'ip ', '{{{', 'ip']
    item_list = re.split('({+[{|#|%])', data)
    if len(item_list) == 1:
        return data

    left_pattern_index = None
    for index, item in enumerate(item_list):
        if index % 2:
            # 1. Handle more than 2 big brackets
            if (len(item) > 2) or (item not in keywords_pair):
                item_list[index] = "{{ '%s' }}" % (item)
            else:
                left_pattern_index = index
            continue

        if left_pattern_index is None:
            continue

        left_pattern = item_list[left_pattern_index]
        left_pattern_index = None

        # 2. Handle mismatched keyword
        right_pattern = keywords_pair[left_pattern]
        if right_pattern not in item:
            item_list[index-1] = "{{ '%s' }}" % (item_list[index-1])
            continue

        # 3. Handle unknown arguments between {{ and }}
        if left_pattern == '{{':
            key_n_lefted = item.split('}}', 1)
            if len(key_n_lefted) != 2:
                continue
            key, _ = key_n_lefted
            key = key.strip()
            if [key for p in params if key.startswith(p)]:
                continue
            item_list[index-1] = "{{ '%s' }}" % (item_list[index-1])

    after_data = ''.join(item_list)
    return after_data


def render(data, enable_tojson=True):
    if not isinstance(data, str):
        logger.warning(f'Format error! Expected str, found {type(data)}')
        return

    params = {
        'config': config,
        'ip': config.get('ip'),
        'port': config.get('mock.port'),
        'today': datetime.date.today(),
        'now':  datetime.datetime.now()
    }

    if enable_tojson:
        data = handle_jinja2_tojson_by_config(data)

    # Jinja2 doc
    # undefined and UndefinedError https://jinja.palletsprojects.com/en/3.1.x/api/#undefined-types
    # TemplateSyntaxError https://jinja.palletsprojects.com/en/3.1.x/api/#jinja2.TemplateSyntaxError

    try:
        template_data = Template(data, keep_trailing_newline=True, undefined=StrictUndefined)
        data = template_data.render(params)
    except (UndefinedError, TemplateSyntaxError):
        data = handle_jinja2_keywords(data, params)
        template_data = Template(data, keep_trailing_newline=True)
        data = template_data.render(params)
    except Exception:
        logger.error(f'Format error!\n {traceback.format_exc()}')

    return data


def get_query_array(url):
    # query string to array, example:
    # a=1&b=2 ==> ['a', '1', 'b', '2']
    # a=&b=2 ==> ['a', '', 'b', '2']
    # a&b=2 ==> ['a', '', 'b', '2']
    query_array = []
    qs_index = url.find('?')
    if qs_index < 0:
        return query_array

    query_string = url[qs_index+1:]
    if not query_string:
        return query_array

    for k_v in query_string.split('&'):
        if not k_v:
            continue

        k_v_item = k_v.split('=')
        if len(k_v_item) >= 2:
            query_array.extend(k_v_item[:2])
        else:
            query_array.extend([k_v, ''])
    return query_array


def url_decode(decode_obj, decode_key):
    if not isinstance(decode_obj, (dict, list)):
        return
    if isinstance(decode_obj, dict) and decode_key not in decode_obj:
        return
    if isinstance(decode_obj, list) and (not isinstance(decode_key, int) or decode_key >= len(decode_obj)):
        return
    url_decode_for_list_or_dict(decode_obj, decode_key)


def url_decode_for_list_or_dict(decode_obj, decode_key):
    if isinstance(decode_obj[decode_key], str):
        decode_obj[decode_key] = unquote(decode_obj[decode_key])
    elif isinstance(decode_obj[decode_key], list):
        for idx, _ in enumerate(decode_obj[decode_key]):
            url_decode(decode_obj[decode_key], idx)
    elif isinstance(decode_obj[decode_key], dict):
        for key, _ in decode_obj[decode_key].items():
            url_decode(decode_obj[decode_key], key)


def flow_str_2_data(data_str):
    if not isinstance(data_str, str):
        return data_str
    try:
        return json.loads(data_str)
    except Exception:
        return data_str


def flow_data_2_str(data):
    if isinstance(data, str):
        return data
    return json.dumps(data, ensure_ascii=False)


class CaseInsensitiveDict(dict):
    '''
    A dict data-structure that ignore key's case.
    Any read or write related operations will igonre key's case.

    Example:
    <key: 'abc'> & <key: 'ABC'> will be treated as the same key, only one exists in this dict.
    '''

    def __init__(self, raw_dict):
        self.__key_map = {}
        for k, v in raw_dict.items():
            self.__setitem__(k, v)

    def __get_real_key(self, key):
        return self.__key_map.get(key.lower(), key)

    def __set_real_key(self, real_key):
        if real_key.lower() not in self.__key_map:
            self.__key_map[real_key.lower()] = real_key

    def __pop_real_key(self, key):
        return self.__key_map.pop(key.lower())

    def __del_real_key(self, key):
        del self.__key_map[key.lower()]

    def __clear_key_map(self):
        self.__key_map.clear()

    def __setitem__(self, key, value):
        real_key = self.__get_real_key(key)
        self.__set_real_key(real_key)
        super(CaseInsensitiveDict, self).__setitem__(real_key, value)

    def __getitem__(self, key):
        return super(CaseInsensitiveDict, self).__getitem__(self.__get_real_key(key))

    def get(self, key, default=None):
        return super(CaseInsensitiveDict, self).get(self.__get_real_key(key), default)

    def __delitem__(self, key):
        real_key = self.__pop_real_key(key)
        return super(CaseInsensitiveDict, self).__delitem__(real_key)

    def __contains__(self, key):
        return key.lower() in self.__key_map

    def pop(self, key):
        real_key = self.__pop_real_key(key)
        return super(CaseInsensitiveDict, self).pop(real_key)

    def popitem(self):
        item = super(CaseInsensitiveDict, self).popitem()
        if item:
            self.__del_real_key(item[0])
        return item

    def clear(self):
        self.__clear_key_map()
        return super(CaseInsensitiveDict, self).clear()

    def update(self, __m=None, **kwargs) -> None:
        if __m:
            for k, v in __m.items():
                self.__setitem__(k, v)
        if kwargs:
            for k, v in kwargs.items():
                self.__setitem__(k, v)


class HookedDict(dict):
    '''
    Hook build-in dict to protect CaseInsensitiveDict data type.
    Only <headers> value is CaseInsensitiveDict at present.
    '''

    def __init__(self, raw_dict):
        for k, v in raw_dict.items():
            if type(v) == dict:
                if k.lower() == 'headers':
                    v = CaseInsensitiveDict(v)
                else:
                    v = HookedDict(v)
            self.__setitem__(k, v)

    def __setitem__(self, __k, __v) -> None:
        if type(__v) == dict:
            if __k.lower() == 'headers':
                __v = CaseInsensitiveDict(__v)
            else:
                __v = HookedDict(__v)
        return super(HookedDict, self).__setitem__(__k, __v)


class TargetMatch:

    @staticmethod
    def is_match(target, pattern):
        if not TargetMatch._match_type(target, pattern):
            return False
        if type(target) == str:
            return TargetMatch._match_string(target, pattern)
        elif type(target) in [int, float]:
            return TargetMatch._match_numbers(target, pattern)
        elif type(target) == bool:
            return TargetMatch._match_boolean(target, pattern)
        elif type(target).__name__ == 'NoneType':
            return TargetMatch._match_null(target, pattern)
        else:
            logger.warning(f'Illegal match target type: {type(target)}')
            return False

    @staticmethod
    def _match_type(target, pattern):
        return isinstance(target, type(pattern))

    @staticmethod
    def _match_string(target, pattern):
        return True if re.search(pattern, target) else False

    @staticmethod
    def _match_numbers(target, pattern):
        return target == pattern

    @staticmethod
    def _match_boolean(target, pattern):
        return target == pattern

    @staticmethod
    def _match_null(target, pattern):
        return target == pattern


class JSONFormat:

    def json(self):
        prop_collection = {}
        props = dir(self)
        for prop in props:
            if prop.startswith('_'):
                continue
            prop_obj = getattr(self, prop)
            if isinstance(prop_obj, (str, int, bool, float)):
                prop_collection[prop] = prop_obj
            elif isinstance(prop_obj, datetime.datetime):
                prop_collection[prop] = prop_obj.timestamp()
        return prop_collection



import uuid
import traceback
from datetime import datetime
from urllib.parse import urlparse
from collections import OrderedDict

from lyrebird import application
from lyrebird.log import get_logger
from lyrebird.mock.dm.match import MatchRules
from lyrebird.utils import flow_data_2_str, render


logger = get_logger()


class TempMock:
    def __init__(self):
        self.root = {
            'id': 'tmp_group',
            'type': 'group',
            'name': 'Temp Mock',
            'label': [],
            'children': []
        }
        self.activated_data = OrderedDict()

    def get(self):
        return self.root
    
    def get_matched_data(self, flow):
        _matched_data = []
        for data in self.activated_data.values():
            if MatchRules.match(flow, data.get('rule')):
                _matched_data.append(data)
                break

        for response_data in _matched_data:
            if 'response' not in response_data:
                continue
            if 'data' not in response_data['response']:
                continue
            if not response_data['response']['data']:
                continue

            origin_response_data = response_data['response']['data']
            try:
                response_data['response']['data'] = render(response_data['response']['data'])
            except Exception:
                response_data['response']['data'] = origin_response_data
                logger.warning(f'Format response data error! {response_data["request"]["url"]}\n {traceback.format_exc()}')

        return _matched_data

    def add_data(self, raw_data):
        if not isinstance(raw_data, dict):
            raise DataObjectShouldBeADict

        data = dict(raw_data)
        data_id = str(uuid.uuid4())
        data['id'] = data_id

        if 'request' in data:
            data['request'] = dict(raw_data['request'])

            _data_name = data['name'] if data.get('name') else self._get_data_name(data)
            _data_rule = data['rule'] if data.get('rule') else self._get_data_rule(data['request'])
            if 'data' in data['request']:
                data['request']['data'] = flow_data_2_str(data['request']['data'])
        else:
            _data_name = data.get('name')
            _data_rule = {'request.url': '(?=.*YOUR-REQUEST-PATH)(?=.*PARAMS)'}
            data['request'] = {}

        data['name'] = _data_name
        data['rule'] = _data_rule

        if 'response' in data:
            data['response'] = dict(raw_data['response'])

            if 'data' in data['response']:
                data['response']['data'] = flow_data_2_str(data['response']['data'])
        else:
            data['response'] = {}

        self.activated_data[data_id] = data
        self.activated_data.move_to_end(data_id, last=False)

        data_node = {
            'id': data_id,
            'name': data.get('name'),
            'type': 'data',
            'parent_id': 'tmp_group'
        }
        self.root['children'].insert(0, data_node)

        return data_id

    def delete_by_query(self, query):
        if not query.get('id'):
            return
        ids = query['id']

        for id_ in ids:
            if id_ in self.activated_data:
                del self.activated_data[id_]

        for child in self.root['children'][::-1]:
            if child['id'] in ids:
                self.root['children'].remove(child)

    @staticmethod
    def _get_data_name(data):
        name = 'New Data'
        request = data.get('request')
        if not request:
            return name

        url = request.get('url')
        if not url:
            return name

        parsed_url = urlparse(url)
        host = parsed_url.hostname
        path = parsed_url.path
        if path:
            name = path
        elif host:
            name = host
        else:
            name = url[0:100]

        time_now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        name = f'{time_now}  {name}'

        return name

    @staticmethod
    def _get_data_rule(request):
        pattern = 'YOUR-REQUEST-PATH'
        url = request.get('url')
        if not url:
            return {'request.url': f'(?=.*{pattern})'}

        parsed_url = urlparse(url)
        host = parsed_url.netloc
        path = parsed_url.path
        query = parsed_url.query
        if path and host:
            pattern = path
            if query:
                pattern += '\?' 
            elif url.endswith(path):
                pattern += '$'

        elif host:
            pattern = host
            if url.endswith(host):
                pattern += '$'

        else:
            pattern = url
        return {'request.url': f'(?=.*{pattern})'}


class DataObjectShouldBeADict(Exception):
    pass



from collections import OrderedDict
from .content_length import ContentLengthHandler
from ..duplicate_header_key_handler import DuplicateHeaderKeyHandler

origin2flow_handlers = OrderedDict({
})

flow2origin_handlers = OrderedDict({
    'Content-Length': ContentLengthHandler
})


class HeadersHelper:

    @staticmethod
    def origin2flow(origin_obj, output=None):
        _origin_headers = origin_obj.headers
        if not _origin_headers:
            return

        _headers = DuplicateHeaderKeyHandler.origin2flow(_origin_headers)

        for headers_key, func in origin2flow_handlers.items():
            _headers[headers_key] = func.flow2origin(origin_obj)

        if output:
            output['headers'] = _headers
        else:
            return _headers

    @staticmethod
    def flow2origin(flow_obj, output=None, chain=None):
        _headers = flow_obj.get('headers')
        if not _headers:
            return

        for headers_key, func in flow2origin_handlers.items():
            _headers[headers_key] = func.flow2origin(flow_obj, chain=chain)

        if output:
            output.headers = _headers
        else:
            return _headers




import json


class JsonHandler:

    @staticmethod
    def origin2flow(request_data):
        _data = json.loads(request_data.decode('utf-8'))
        return _data

    @staticmethod
    def flow2origin(flow_data):
        _data = json.dumps(flow_data, ensure_ascii=False).encode()
        return _data

    @staticmethod
    def origin2string(request_data):
        return request_data.decode('utf-8')



from flask_restful import Resource
from flask import request
from lyrebird.mock import context, headers
from lyrebird import application
from lyrebird.event_filter import Filter
from lyrebird.application import config
from urllib.parse import urlencode, unquote
from flask import request, Response
from copy import deepcopy
from lyrebird.utils import url_decode
import json


class Flow(Resource):
    """
    当前请求单条数据
    """

    def get(self, id):
        is_origin = request.args.get('is_origin', 'false').strip().lower() == 'true'
        no_decode = request.args.get('no_decode')
        if no_decode is None:
            no_decode = config.get('inspector.detail.no_decode', 0)
        if isinstance(no_decode, str) and no_decode.strip().isdigit():
            no_decode = int(no_decode)
        elif not isinstance(no_decode, int):
            return application.make_fail_response(f'Param no_decode type error, in the request of api/flow/{id}')
        
        for item in context.application.cache.items():
            if item['id'] == id:
                # Import decoder for decoding the requested content
                display_item = {}
                if is_origin:
                    display_item = deepcopy(item)
                else:
                    application.encoders_decoders.decoder_handler(item, output=display_item)
                if not no_decode:
                    for key in ('url', 'path', 'query'):
                        url_decode(display_item['request'], key)
                return application.make_ok_response(data=display_item)
        return application.make_fail_response(f'Request {id} not found')


def get_flow_list_by_filter(filter_obj):
    all_items = context.application.cache.items()[::-1]
    req_list = []
    target_items = Filter.get_items_after_filtration(all_items, filter_obj)
    for item in target_items:
        info = dict(
            id=item['id'],
            size=item['size'],
            duration=item['duration'],
            start_time=item['start_time'],
            request=dict(
                url=item['request'].get('url'),
                scheme=item['request'].get('scheme'),
                host=item['request'].get('host'),
                path=item['request'].get('path'),
                params=unquote(urlencode(item['request']['query'])),
                method=item['request'].get('method')
            ),
            response=dict(
                code=item['response']['code'],
                mock=item['response']['headers'].get('lyrebird', 'proxy'),
                modified=item['request']['headers'].get('lyrebird_modified') or item['response']['headers'].get('lyrebird_modified', '')
            )if item.get('response') and len(item.get('response').keys()) > 1 else {},
            action=item.get('action', [])
        )
        # Add key `request.port` when port is not default
        if item['request'].get('port') not in ['443', '80']:
            info['request']['port'] = item['request'].get('port')

        # Add key `proxy_response` into info only if item contains proxy_response
        if item.get('proxy_response'):
            info['proxy_response'] = {
                'code': item['proxy_response']['code']
            }
        # Change status
        if item['request']['headers'].get(headers.MITMPROXY_COMMAND):
            info['status'] = item['request']['headers'][headers.MITMPROXY_COMMAND]
        # SSR
        if item['request']['headers'].get('lyrebird') == 'mock':
            info['status'] = 'ssr'

        req_list.append(info)
    return req_list


class FlowList(Resource):
    """
    当前请求列表
    """

    def get(self):
        default_filter = context.application.selected_filter
        req_list = get_flow_list_by_filter(default_filter)
        return Response(json.dumps(req_list, ensure_ascii=False), mimetype='application/json', status=200)

    def delete(self):
        _ids = request.json.get('ids')
        if _ids:
            context.application.cache.delete_by_ids(*_ids)
        else:
            context.application.cache.clear()
        context.application.socket_io.emit('action', 'delete flow log')
        return application.make_ok_response()

    def post(self, action):
        if action == 'save':
            _ids = request.json.get('ids')
            record_items = []
            for _id in _ids:
                for item in context.application.cache.items():
                    if _id == item['id']:
                        record_items.append(item)
                        break
            dm = context.application.data_manager

            for flow in record_items:
                dm.save_data(flow)
            return application.make_ok_response()
        elif action == 'search':
            filter_name = request.json.get('selectedFilter')
            filter_obj = context.application.get_and_update_selected_filter_by_name(filter_name)
            req_list = get_flow_list_by_filter(filter_obj)
            return Response(json.dumps(req_list, ensure_ascii=False), mimetype='application/json', status=200)
        else:
            return application.make_fail_response(f'action: {action} is not supported')



from flask_restful import Resource
from lyrebird.mock import context
from lyrebird import application


class ConflictCheck(Resource):

    def get(self, group_id):
        result = context.application.data_manager.check_conflict(group_id)
        return application.make_ok_response(data=result)


class ActivatedDataConflictCheck(Resource):

    def get(self):
        result = context.application.data_manager.activated_data_check_conflict()
        return application.make_ok_response(data=result)



from lyrebird.mock import context


class SettingDiffMode:

    @staticmethod
    def add(mode):
        context.application.is_diff_mode = mode

    @staticmethod
    def remove(mode):
        context.application.is_diff_mode = mode



from flask import Blueprint, request, Response

from ..handlers.mock_handler import MockHandler
from ..handlers.proxy_handler import ProxyHandler
from ..handlers.duplicate_header_key_handler import DuplicateHeaderKeyHandler
from ..handlers.path_not_found_handler import RequestPathNotFound
from ..handlers.handler_context import HandlerContext
from ..handlers.flow_editor_handler import FlowEditorHandler
from .. import context
from lyrebird import log


logger = log.get_logger()
mock_handler = MockHandler()
proxy_handler = ProxyHandler()
path_not_found_handler = RequestPathNotFound()
flow_editor_handler = FlowEditorHandler()


core = Blueprint('mock', __name__, url_prefix='/mock')


@core.route('/', methods=['GET', 'POST', 'HEAD', 'PUT', 'DELETE', 'OPTIONS'])
@core.route('/<path:path>', methods=['GET', 'POST', 'HEAD', 'PUT', 'DELETE', 'OPTIONS'])
def index(path=''):
    logger.info(f'<Core> On request {request.url}')

    resp = None
    req_context = HandlerContext(request)
    req_context.update_client_req_time()

    flow_editor_handler.on_request_handler(req_context)

    req_context.update_server_req_time()

    mock_handler.handle(req_context)

    if not req_context.response_source:
        flow_editor_handler.on_request_upstream_handler(req_context)
        proxy_handler.handle(req_context)
        if req_context.is_proxiable:
            req_context.set_response_source_proxy()
            req_context.update_response_headers_code2flow()
            flow_editor_handler.on_response_upstream_handler(req_context)

    req_context.update_server_resp_time()

    flow_editor_handler.on_response_handler(req_context)

    if req_context.response_source:
        gen = req_context.get_response_generator()
        resp = Response(
            gen(),
            status=req_context.flow['response']['code'],
            headers=req_context.flow['response']['headers']
        )
    elif req_context.flow['response']:
        resp = Response(
            req_context.flow['response'].get('data', ''),
            status=req_context.flow['response'].get('code', 200),
            headers=req_context.flow['response'].get('headers', {})
        )
    else:
        path_not_found_handler.handle(req_context)
        req_context.update_client_resp_time()
        resp = req_context.response

    DuplicateHeaderKeyHandler.set_origin_header(resp.headers, req_context.flow['response']['headers'])

    context.emit('action', 'add flow log')

    # Close autocorrect response location.
    resp.autocorrect_location_header = False

    return resp



import json
from io import BytesIO
from lyrebird.mock.dm.jsonpath import jsonpath


class EventFileConverter:
    
    def __init__(self, data, target_node_path=None):
        self.data = data
        self.target_node_path = target_node_path
        self.suffix = 'json'

    @property
    def filename(self):
        return f'{self.data["channel"]}_{self.data["id"]}.{self.suffix}'
    
    def convert(self):
        pass


class JSONFileNConverter(EventFileConverter):

    def __init__(self, data, target_node_path=None):
        super().__init__(data, target_node_path)
        self.suffix = 'json'
    
    def convert(self):
        # Covert common json to bytes
        target_nodes = jsonpath.search(self.data, self.target_node_path)
        if not target_nodes:
            raise ConvertDataToStreamFail('Cannot find target path: {self.target_node_path}')
        target_json = target_nodes[0].node
        try:
            return bytes(json.dumps(target_json, indent=4, ensure_ascii=False), encoding='utf-8')
        except Exception as e:
            raise ConvertDataToStreamFail(e)


class FlowJSONFileConverter(JSONFileNConverter):

    def __init__(self, data):
        super().__init__(data, 'flow')


class ConvertDataToStreamFail(Exception):
    pass


class ConverterNotSupport(Exception):
    pass


EVENT_CONVERTER_MAP = {
    'flow_json': FlowJSONFileConverter,
    'json': JSONFileNConverter
}


def get_event_converter(event):
    converter = event.get('export').get('converter')
    if not converter:
        raise ConverterNotSupport('Converter is None')
    event_converter = EVENT_CONVERTER_MAP.get(converter)
    if not event_converter:
        raise ConverterNotSupport(f'Converter "{converter}" is not supported.')
    return event_converter(event)


def _generator_export_stream(data_bytes):
    def generator():
        file_like_io = BytesIO(data_bytes)
        for item in file_like_io.readlines():
            yield item
        file_like_io.close()
    return generator


def export_from_event(event):
    converter = get_event_converter(event)
    filename = converter.filename
    data_bytes = converter.convert()
    return filename, _generator_export_stream(data_bytes)
     